import 'package:dart_arango_min/src/dart_arango_min_base.dart' as base;

abstract class QueryTextFragment {
  String line;
  String toString() => line;
}

/// Simple text line for injecting it into the query, constructed by [Query()].
/// Sample:
///  ```
/// Query([ Line('FOR doc in documents'), ... ])
/// ```
class Line extends QueryTextFragment {
  String line;

  Line(this.line);

  @override
  String toString() => line;
}

/// Line of query text with condition.
/// If [bool] `cond` is true, then [String] `when_true` will pasted into query.
/// Sample:
/// ```
/// Query([
///   Line('FOR doc in documents'),
///   LineIfThen(tag != null, 'FILTER doc.tags && POSITION( doc.tags, tag )'),
///   ...
/// ])
/// ```
class LineIfThen extends QueryTextFragment {
  LineIfThen(bool cond, String when_true) {
    line = cond ? when_true : '';
  }
}

/// Line of query text with condition.
/// If [bool] `cond` is true, then [String] `when_true` will pasted into query.
/// Else [String] `when_false` willpasted into query.
/// Sample:
/// ```
/// Query([
///   Line('FOR doc in documents'),
///   LineIfThen(tag != null, 'FILTER doc.tags && POSITION( doc.tags, tag )'),
///   Line('SORT doc.datetime'),
///   LineIfThenElse( without_content, 'return UNSET( doc, "content" )', 'return doc'),
///   ...
/// ])
/// ```
class LineIfThenElse extends QueryTextFragment {
  LineIfThenElse(bool cond, String when_true, String when_false) {
    line = cond ? when_true : when_false;
  }
}

abstract class BindNameValuePair {
  Map<String, dynamic> bindNameValuePair = {};
}

/// Simple pair of name-value to bind they to query as variable.
/// Sample:
/// ```
/// Query(
///      [
///         Line('LET tag=@tag'),
///         ... list of query lines (see Line(), LineIfThen(), IfThemElse() for details )
///      ],
///      bindVars: [ BindVar( 'tag', tag ) ],
///    ).toMap()
/// ```
class BindVar extends BindNameValuePair {
  BindVar(String name, value) {
    bindNameValuePair = {name: value};
  }
}

/// Pair of variable name-value to bind they into query with condition.
/// Sample:
/// ```
/// Query(
///      [
///         LineIfThen( tag != null, 'LET tag=@tag'),
///         // ... list of other query lines (see Line(), LineIfThen(), LineIfThenElse() for details )
///      ],
///      bindVars: [ BindVarIfThen( tag!=null, 'tag', tag ) ],
///    ).toMap()
/// ```
class BindVarIfThen extends BindNameValuePair {
  BindVarIfThen(bool condition, String name_if_true, value_if_true) {
    if (condition) {
      this.bindNameValuePair = {name_if_true: value_if_true};
    }
  }
}

/// Pair of variable name-value to bind they to query with condition.
/// Sample:
/// ```
/// Query(
///      [
///         LineIfThenElse( tag != null, 'LET tag=@tag', 'LET other_name=@other_name'),
///         // ... list of query lines (see Line(), LineIfThen(), IfThemElse() for details )
///      ],
///      bindVars: [ BindVarIfThenElse( tag!=null, 'tag', tag, 'other_name', other_value  ) ],
///    ).toMap()
/// ```
class BindVarIfThenElse extends BindNameValuePair {
  BindVarIfThenElse(bool condition, String name_if_true, value_if_true,
      String name_if_false, value_if_false) {
    if (condition) {
      this.bindNameValuePair = {name_if_true: value_if_true};
    } else {
      this.bindNameValuePair = {name_if_false: value_if_false};
    }
  }
}

/// Constructor for constructing query as object from understandable arguments with condition support.
/// Query object do not call database.
/// Query object can be converting to [Map] via toMap() method and used as argument to `query()` method 
/// of database client for sending.
/// For debug purpose query object can be printed as [String], because it has toString() method.
/// Sample:
/// ```
/// var query = Query(
///      [
///        Line('LET tag=@tag'),
///        Line('FOR doc in docs'),
///        LineIfThen(tag != null, 'FILTER doc.tags && POSITION( doc.tags, tag )'),
///        Line('SORT doc.datetime'),
///        LineIfThenElse(without_content, 'return UNSET(doc,"content")', 'return doc'),
///      ],
///      bindVars: [BindVarIfThen(tag != null, 'tag', tag)],
///    ).toMap();
/// ```
class Query {

  // исходные структуры, задаются в конструкторах:
  List<QueryTextFragment> _fragments = [];
  List<BindNameValuePair> _bindVars = [];

  // Результирующий собранный из всех _bindVars:
  Map<String, dynamic> _bindVarsMap = {};

  /// Returns binded vars from query object.
  get bindedVars => this._bindVarsMap;

  Query(this._fragments, {List<BindNameValuePair> bindVars}) {
    this._bindVars = bindVars;
  }

  /// Like Query() but named constructor.
  Query.Create( fragments, {List<BindNameValuePair> bindVars} ){
    Query( fragments, bindVars: bindVars);
  }

  /// Adds one [QueryTextFragment] to inner store of query fragments in during step by step constructing Query.
  /// 
  /// Sample:
  /// ```
  /// .addQueryTextFragment( Line( 'FOR u IN users RETURN u' ) )
  /// ```
  Query addQueryTextFragment( QueryTextFragment fragment ){
    this._fragments.add(fragment);
    return this;
  }

  /// Adds one [Line] to inner store of query fragments.
  /// 
  /// Sample:
  /// ```
  /// .addLine( 'FOR u IN users RETURN u' )
  /// ```
  Query addLine( String line ) => addQueryTextFragment( Line( line ) );

  /// Adds one [LineIfThen] to inner store of query fragments.
  /// 
  /// Sample:
  /// ```
  /// .addLineIfThen( user_name!=null, 'FILTER u.name==@user_name' )
  /// ```
  Query addLineIfThen( bool cond, String line ) => addQueryTextFragment( LineIfThen( cond, line ) );

  /// Adds one pair of name-value binded to query
  /// Sample:
  /// ```
  /// .addBindNameValuePair( BindVar( 'user_name', 'Dmitriy' ) )
  /// ```
  Query addBindNameValuePair( BindNameValuePair name_value_pair ){
    this._bindVars.add( name_value_pair );
    return this;
  }

  /// Adds one pair of name-value binded to query
  /// Sample:
  /// ```
  /// .addBindVar( 'user_name', 'Dmitry' )
  /// ```
  Query addBindVar( String name, value ) => addBindNameValuePair( BindVar( name, value ) );


  /// Adds one condition-depended pair of name-value binded to query
  /// Sample:
  /// ```
  /// .addBindVarIfThen( user_name!=null, 'user_name', user_name )
  /// ```
  Query addBindVarIfThen( bool cond, String name, value ) => addBindNameValuePair(BindVarIfThen(cond, name, value));

  /// Adds one condition-depended pair of name-value binded to query
  /// Sample:
  /// ```
  /// .addBindVarIfThenElse( user_name!=null, 'user_name', user_name, 'max_user_age', 30 )
  /// ```
  Query addBindVarIfThenElse( bool cond, String name_if_true, value_if_true, name_if_false, values_if_false ) => 
    addBindNameValuePair(BindVarIfThenElse(cond, name_if_true, value_if_true, name_if_false, values_if_false ));

  /// Returns query string from [Query] object.
  String queryString() => this
      ._fragments
      .map((f) => f.toString())
      .where((f) => f.isNotEmpty)
      .toList()
      .cast<String>()
      .join('\n');

  /// Returns created query object as [Map] structure, which contains keys: `'query'` and `'bindVars'`.
  /// The `'query'` key value contains a query text, `queryString()` method result.
  /// The `'bindVars'` key value contains a structure with binded variables, `bindedVars` getter result.  
  Map<String,dynamic> toMap() {
    //соберем bindVars в один Map:
    for (var bv in this._bindVars) {
      this._bindVarsMap.addAll(bv.bindNameValuePair);
    }

    var result = {
      'query': queryString(),
      'bindVars': bindedVars,
    };

    return result;
  }

  /// Human readable view of Query object.
  /// See `toMap()` method to get created query object as Map structure.
  @override
  toString() => toMap().toString();

}

class QueryWithClient extends Query {
  // ссылка на объект, создавший данный запрос.
   base.ArangoDBClient client;

  QueryWithClient( this.client, List<QueryTextFragment> fragments, {List<BindNameValuePair> bindVars} )
    : super.Create( fragments, bindVars: bindVars );
  
  /// Calls `client.queryToList( this.toMap() )`
  /// where `client` is [ArangoDBClient].
  RunAndReturnFutureList() async => await client.queryToList( this.toMap() );

  RunAndReturnStream() async => await client.queryToStream( this.toMap() );
}

